from typing import Generator, Any

from django.contrib import admin
from django.contrib.admin import ModelAdmin
from django.contrib.admin.views.main import ChangeList
from django.core.exceptions import ValidationError
from django.core.validators import EMPTY_VALUES
from django.db.models import QuerySet, Model, Field, DateField, DateTimeField
from django.http import HttpRequest
from django.utils.translation import gettext_lazy as _

from dashub.contrib.forms import RadioForm, CheckboxForm, DropdownForm, AutocompleteDropdownForm, RangeDateForm, \
    RangeDateTimeForm
from dashub.contrib.mixins import ValueMixin, ChoicesMixin, MultiValueMixin, DropdownMixin, AutocompleteMixin
from dashub.utils import parse_date_str, parse_datetime_str


class RadioFilter(admin.SimpleListFilter):
    template = "dashub/filters/filters_field.html"
    form_class = RadioForm
    all_option = ["", _("All")]

    def choices(self, changelist: ChangeList) -> tuple[dict[str, Any], ...]:
        add_facets = getattr(changelist, "add_facets", False)
        facet_counts = self.get_facet_queryset(changelist) if add_facets else None
        choices = []

        if self.all_option:
            choices = [self.all_option]

        if add_facets:
            for i, (lookup, title) in enumerate(self.lookup_choices):
                if (count := facet_counts.get(f"{i}__c", -1)) != -1:
                    title = f"{title} ({count})"
                else:
                    title = f"{title} (-)"

                choices.append((lookup, title))
        else:
            choices.extend(self.lookup_choices)

        return (
            {
                "form": self.form_class(
                    label=_(" By %(filter_title)s ") % {"filter_title": self.title},
                    name=self.parameter_name,
                    choices=choices,
                    data={self.parameter_name: self.value()},
                ),
            },
        )


class CheckboxFilter(RadioFilter):
    form_class = CheckboxForm
    all_option = None

    def value(self) -> list[Any]:
        return self.request.GET.getlist(self.parameter_name)


class ChoicesRadioFilter(ValueMixin, ChoicesMixin, admin.ChoicesFieldListFilter):
    form_class = RadioForm
    all_option = ["", _("All")]


class ChoicesCheckboxFilter(
    MultiValueMixin, ChoicesMixin, admin.ChoicesFieldListFilter
):
    form_class = CheckboxForm
    all_option = None


class BooleanRadioFilter(ValueMixin, admin.BooleanFieldListFilter):
    template = "dashub/filters/filters_field.html"
    all_option = ["", _("All")]
    form_class = RadioForm

    def choices(self, changelist: ChangeList) -> Generator[dict[str, Any], None, None]:
        add_facets = getattr(changelist, "add_facets", False)
        facet_counts = self.get_facet_queryset(changelist) if add_facets else None

        if add_facets:
            choices = [
                self.all_option,
                *[
                    ("1", f"{_('Yes')} ({facet_counts['true__c']})"),
                    ("0", f"{_('No')} ({facet_counts['false__c']})"),
                ],
            ]
        else:
            choices = [
                self.all_option,
                *[
                    ("1", _("Yes")),
                    ("0", _("No")),
                ],
            ]

        yield {
            "form": self.form_class(
                label=_(" By %(filter_title)s ") % {"filter_title": self.title},
                name=self.lookup_kwarg,
                choices=choices,
                data={self.lookup_kwarg: self.value()},
            ),
        }


class RelatedCheckboxFilter(MultiValueMixin, admin.RelatedFieldListFilter):
    template = "dashub/filters/filters_field.html"
    form_class = CheckboxForm

    def queryset(self, request: HttpRequest, queryset: QuerySet) -> QuerySet:
        if self.value() not in EMPTY_VALUES:
            return super().queryset(request, queryset)
        return queryset

    def choices(self, changelist: ChangeList) -> Generator[dict[str, Any], None, None]:
        add_facets = getattr(changelist, "add_facets", False)
        facet_counts = self.get_facet_queryset(changelist) if add_facets else None

        if add_facets:
            choices = []

            for pk_val, val in self.lookup_choices:
                count = facet_counts[f"{pk_val}__c"]
                choice = (pk_val, f"{val} ({count})")
                choices.append(choice)
        else:
            choices = self.lookup_choices

        yield {
            "form": self.form_class(
                label=_(" By %(filter_title)s ") % {"filter_title": self.title},
                name=self.lookup_kwarg,
                choices=choices,
                data={self.lookup_kwarg: self.value()},
            ),
        }



class DropdownFilter(admin.SimpleListFilter):
    template = "dashub/filters/filters_field.html"
    form_class = DropdownForm
    all_option = ["", _("All")]

    def choices(self, changelist: ChangeList) -> tuple[dict[str, Any], ...]:
        return (
            {
                "form": self.form_class(
                    label=_(" By %(filter_title)s ") % {"filter_title": self.title},
                    name=self.parameter_name,
                    choices=[self.all_option, *self.lookup_choices],
                    data={self.parameter_name: self.value()},
                    multiple=self.multiple if hasattr(self, "multiple") else False,
                ),
            },
        )


class MultipleDropdownFilter(DropdownFilter):
    multiple = True

    def __init__(
        self,
        request: HttpRequest,
        params: dict[str, Any],
        model: type[Model],
        model_admin: ModelAdmin,
    ) -> None:
        self.request = request
        super().__init__(request, params, model, model_admin)

    def value(self) -> list[Any]:
        return self.request.GET.getlist(self.parameter_name)


class ChoicesDropdownFilter(ValueMixin, DropdownMixin, admin.ChoicesFieldListFilter):
    def choices(self, changelist: ChangeList) -> Generator[dict[str, Any], None, None]:
        choices = [self.all_option, *self.field.flatchoices]

        yield {
            "form": self.form_class(
                label=_(" By %(filter_title)s ") % {"filter_title": self.title},
                name=self.lookup_kwarg,
                choices=choices,
                data={self.lookup_kwarg: self.value()},
                multiple=self.multiple if hasattr(self, "multiple") else False,
            ),
        }


class ChoicesMultipleDropdownFilter(MultiValueMixin, ChoicesDropdownFilter):
    multiple = True


class RelatedDropdownFilter(ValueMixin, DropdownMixin, admin.RelatedFieldListFilter):
    def __init__(
        self,
        field: Field,
        request: HttpRequest,
        params: dict[str, str],
        model: type[Model],
        model_admin: ModelAdmin,
        field_path: str,
    ) -> None:
        super().__init__(field, request, params, model, model_admin, field_path)
        self.model_admin = model_admin
        self.request = request

    def choices(self, changelist: ChangeList) -> Generator[dict[str, Any], None, None]:
        add_facets = getattr(changelist, "add_facets", False)
        facet_counts = self.get_facet_queryset(changelist) if add_facets else None

        if add_facets:
            choices = [self.all_option]

            for pk_val, val in self.lookup_choices:
                if add_facets:
                    count = facet_counts[f"{pk_val}__c"]
                    choice = (pk_val, f"{val} ({count})")
                    choices.append(choice)
        else:
            choices = [self.all_option, *self.lookup_choices]

        yield {
            "form": self.form_class(
                label=_(" By %(filter_title)s ") % {"filter_title": self.title},
                name=self.lookup_kwarg,
                choices=choices,
                data={self.lookup_kwarg: self.value()},
                multiple=self.multiple if hasattr(self, "multiple") else False,
            ),
        }


class MultipleRelatedDropdownFilter(MultiValueMixin, RelatedDropdownFilter):
    multiple = True


class AutocompleteSelectFilter(AutocompleteMixin, RelatedDropdownFilter):
    form_class = AutocompleteDropdownForm


class AutocompleteSelectMultipleFilter(
    AutocompleteMixin, MultipleRelatedDropdownFilter
):
    form_class = AutocompleteDropdownForm



class RangeDateFilter(admin.FieldListFilter):
    request = None
    parameter_name = None
    form_class = RangeDateForm
    template = "dashub/filters/filters_date_range.html"

    def __init__(
        self,
        field: Field,
        request: HttpRequest,
        params: dict[str, str],
        model: type[Model],
        model_admin: ModelAdmin,
        field_path: str,
    ) -> None:
        super().__init__(field, request, params, model, model_admin, field_path)
        if not isinstance(field, DateField):
            raise TypeError(
                f"Class {type(self.field)} is not supported for {self.__class__.__name__}."
            )

        self.request = request
        if self.parameter_name is None:
            self.parameter_name = self.field_path

        if self.parameter_name + "_from" in params:
            value = params.pop(self.field_path + "_from")
            value = value[0] if isinstance(value, list) else value

            if value not in EMPTY_VALUES:
                self.used_parameters[self.field_path + "_from"] = value

        if self.parameter_name + "_to" in params:
            value = params.pop(self.field_path + "_to")
            value = value[0] if isinstance(value, list) else value

            if value not in EMPTY_VALUES:
                self.used_parameters[self.field_path + "_to"] = value

    def queryset(self, request: HttpRequest, queryset: QuerySet) -> QuerySet | None:
        filters = {}

        value_from = self.used_parameters.get(self.parameter_name + "_from")
        if value_from not in EMPTY_VALUES:
            filters.update({self.parameter_name + "__gte": parse_date_str(value_from)})

        value_to = self.used_parameters.get(self.parameter_name + "_to")
        if value_to not in EMPTY_VALUES:
            filters.update({self.parameter_name + "__lte": parse_date_str(value_to)})

        try:
            return queryset.filter(**filters)
        except (ValueError, ValidationError):
            return None

    def expected_parameters(self) -> list[str]:
        return [
            f"{self.parameter_name}_from",
            f"{self.parameter_name}_to",
        ]

    def choices(self, changelist: ChangeList) -> tuple[dict[str, Any], ...]:
        return (
            {
                "request": self.request,
                "parameter_name": self.parameter_name,
                "form": self.form_class(
                    name=self.parameter_name,
                    data={
                        self.parameter_name + "_from": self.used_parameters.get(
                            self.parameter_name + "_from", None
                        ),
                        self.parameter_name + "_to": self.used_parameters.get(
                            self.parameter_name + "_to", None
                        ),
                    },
                ),
            },
        )


class RangeDateTimeFilter(admin.FieldListFilter):
    request = None
    parameter_name = None
    template = "dashub/filters/filters_datetime_range.html"
    form_class = RangeDateTimeForm

    def __init__(
        self,
        field: Field,
        request: HttpRequest,
        params: dict[str, str],
        model: type[Model],
        model_admin: ModelAdmin,
        field_path: str,
    ) -> None:
        super().__init__(field, request, params, model, model_admin, field_path)
        if not isinstance(field, DateTimeField):
            raise TypeError(
                f"Class {type(self.field)} is not supported for {self.__class__.__name__}."
            )

        self.request = request
        if self.parameter_name is None:
            self.parameter_name = self.field_path

        if self.parameter_name + "_from_0" in params:
            value = params.pop(self.field_path + "_from_0")
            value = value[0] if isinstance(value, list) else value
            self.used_parameters[self.field_path + "_from_0"] = value

        if self.parameter_name + "_from_1" in params:
            value = params.pop(self.field_path + "_from_1")
            value = value[0] if isinstance(value, list) else value
            self.used_parameters[self.field_path + "_from_1"] = value

        if self.parameter_name + "_to_0" in params:
            value = params.pop(self.field_path + "_to_0")
            value = value[0] if isinstance(value, list) else value
            self.used_parameters[self.field_path + "_to_0"] = value

        if self.parameter_name + "_to_1" in params:
            value = params.pop(self.field_path + "_to_1")
            value = value[0] if isinstance(value, list) else value
            self.used_parameters[self.field_path + "_to_1"] = value

    def expected_parameters(self) -> list[str]:
        return [
            f"{self.parameter_name}_from_0",
            f"{self.parameter_name}_from_1",
            f"{self.parameter_name}_to_0",
            f"{self.parameter_name}_to_1",
        ]

    def queryset(self, request: HttpRequest, queryset: QuerySet) -> QuerySet | None:
        filters = {}

        date_value_from = self.used_parameters.get(self.parameter_name + "_from_0")
        time_value_from = self.used_parameters.get(self.parameter_name + "_from_1")

        date_value_to = self.used_parameters.get(self.parameter_name + "_to_0")
        time_value_to = self.used_parameters.get(self.parameter_name + "_to_1")

        if date_value_from not in EMPTY_VALUES and time_value_from not in EMPTY_VALUES:
            filters.update(
                {
                    f"{self.parameter_name}__gte": parse_datetime_str(
                        f"{date_value_from} {time_value_from}"
                    ),
                }
            )

        if date_value_to not in EMPTY_VALUES and time_value_to not in EMPTY_VALUES:
            filters.update(
                {
                    f"{self.parameter_name}__lte": parse_datetime_str(
                        f"{date_value_to} {time_value_to}"
                    ),
                }
            )

        try:
            return queryset.filter(**filters)
        except (ValueError, ValidationError):
            return None

    def choices(self, changelist: ChangeList) -> tuple[dict[str, Any], ...]:
        return (
            {
                "request": self.request,
                "parameter_name": self.parameter_name,
                "form": self.form_class(
                    name=self.parameter_name,
                    data={
                        self.parameter_name + "_from_0": self.used_parameters.get(
                            self.parameter_name + "_from_0"
                        ),
                        self.parameter_name + "_from_1": self.used_parameters.get(
                            self.parameter_name + "_from_1"
                        ),
                        self.parameter_name + "_to_0": self.used_parameters.get(
                            self.parameter_name + "_to_0"
                        ),
                        self.parameter_name + "_to_1": self.used_parameters.get(
                            self.parameter_name + "_to_1"
                        ),
                    },
                ),
            },
        )











import copy
import logging
from typing import Any, Dict

from django.conf import settings
from django.core.handlers.wsgi import WSGIRequest

from .utils import hex_to_rgb, evaluate_dynamic_value

logger = logging.getLogger(__name__)

DEFAULT_SETTINGS: Dict[str, Any] = {
    # title of the window (Will default to current_admin_site.site_title)
    "site_title": lambda: "Admin Dashboard",
    # Title on the login screen (19 chars max) (will default to current_admin_site.site_header)
    "site_header": lambda: None,
    # Title on the brand (19 chars max) (will default to current_admin_site.site_header)
    "site_brand": lambda: None,
    # URL of the logo for your site, used for brand on top left
    "site_logo": lambda: "https://cdn.practet.com/static/logo-new.webp",

    "logo_width": None,

    "include_dashboard": True,

    # Relative path to a favicon for your site, will default to site_logo if absent (ideally 32x32 px)
    "site_icon": lambda: None,
    ############
    # Top Menu #
    ############
    # Links to put along the nav bar
    "topmenu_links": [],
    #############
    # User Menu #
    #############
    # Hide these apps when generating side menu e.g (auth)
    "hide_apps": [],
    # Hide these models when generating side menu (e.g auth.user)
    "hide_models": [],
    # List of apps to base side menu ordering off of
    "order_menus": [],
    # Custom links to append to side menu app groups, keyed on lower case app label
    # or makes a new group if the given app label doesnt exist in installed apps
    "custom_links": {},
    # Custom icons for side menu apps/models See the link below
    "icons": {"auth": "fas fa-users-cog", "auth.user": "fas fa-user", "auth.Group": "fas fa-users"},
    # Icons that are used when one is not manually specified
    "default_icon_parents": "fas fa-chevron-circle-right",
    "default_icon_children": "fas fa-circle",
    #############
    # UI Tweaks #
    #############
    # Relative paths to custom CSS/JS scripts (must be present in static files)
    "custom_css": [],
    "custom_js": [],
    ###############
    # Change view #
    ###############
    # Render out the change view as a single form, or in tabs, current options are
    # - single
    # - horizontal_tabs (default)
    # - vertical_tabs
    # - collapsible
    # - carousel
    "changeform_format": "horizontal_tabs",
    # override change forms on a per modeladmin basis
    "changeform_format_overrides": {},
    # Add a language dropdown into the admin
    "language_chooser": False,
    "model_submenus": {},
    "related_modal_active": True,
    "use_compressed_assets": False,

    # Styles
    "theme_color": lambda: "#e31837",
    "border_radius": "4px",

    # Dashboard Overview
    "theme": "system",
    "include_dashboard_list": True,
    "analytics_template": None
}

CHANGEFORM_TEMPLATES = {
    "single": "dashub/includes/single.html",
    "horizontal_tabs": "dashub/includes/horizontal_tabs.html",
    "vertical_tabs": "dashub/includes/vertical_tabs.html",
}


def get_settings() -> Dict:
    dashub_settings = copy.deepcopy(DEFAULT_SETTINGS)
    user_settings = {x: y for x, y in getattr(settings, "DASHUB_SETTINGS", {}).items() if y is not None}
    dashub_settings.update(user_settings)

    # Deal with single strings in hide_apps/hide_models and make sure we lower case 'em
    if isinstance(dashub_settings["hide_apps"], str):
        dashub_settings["hide_apps"] = [dashub_settings["hide_apps"]]
    dashub_settings["hide_apps"] = [x.lower() for x in dashub_settings["hide_apps"]]

    if isinstance(dashub_settings["hide_models"], str):
        dashub_settings["hide_models"] = [dashub_settings["hide_models"]]
    dashub_settings["hide_models"] = [x.lower() for x in dashub_settings["hide_models"]]

    # Ensure icon model names and classes are lower case
    dashub_settings["icons"] = {x.lower(): y.lower() for x, y in dashub_settings.get("icons", {}).items()}

    # Default the site icon using the site logo
    dashub_settings["site_icon"] = dashub_settings["site_icon"] or dashub_settings["site_logo"]

    # ensure all model names are lower cased
    dashub_settings["changeform_format_overrides"] = {
        x.lower(): y.lower() for x, y in dashub_settings.get("changeform_format_overrides", {}).items()
    }

    dashub_settings["theme_color_rgb"] = hex_to_rgb(dashub_settings.get("theme_color", "#30AA99"))
    return dashub_settings


def get_resolved_settings(request: WSGIRequest = None) -> Dict:
    """
    Get settings with all dynamic values resolved
    """
    settings_dict = get_settings()

    for key in settings_dict:
        settings_dict[key] = evaluate_dynamic_value(settings_dict[key], request)

    if 'theme_color' in settings_dict and settings_dict['theme_color']:
        settings_dict["theme_color_rgb"] = hex_to_rgb(settings_dict.get("theme_color", "#30AA99"))

    if not settings_dict.get("site_icon") and settings_dict.get("site_logo"):
        settings_dict["site_icon"] = settings_dict["site_logo"]

    return settings_dict



from django.db import models
from .widgets import TagInputWidget, ColorisWidget
from dashub import widgets as dashub_widgets
from django.contrib.admin import widgets as admin_widgets
from django.forms import fields as form_fields


class TagInputField(models.TextField):
    """
    Custom ModelField that stores a string of tags separated by a custom separator,
    but presents and accepts data as a list in Python.
    """

    def __init__(self, *args, separator=":::", **kwargs):
        self.separator = separator
        super().__init__(*args, **kwargs)

    def formfield(self, **kwargs):
        """Use custom widget for both regular and admin forms"""
        widget = kwargs.get("widget", TagInputWidget)

        if widget == admin_widgets.AdminTextareaWidget:
            widget = dashub_widgets.AdminTagInputWidget(separator=self.separator)

        kwargs["widget"] = widget
        return super().formfield(**kwargs)

    def from_db_value(self, value, expression, connection):
        if not value:
            return []
        return value.split(self.separator)

    def to_python(self, value):
        if value is None:
            return []
        if isinstance(value, list):
            return value
        value = value.strip()
        return value.split(self.separator) if value else []

    def get_prep_value(self, value):
        if value is None:
            return ""
        if isinstance(value, list) or isinstance(value, dict):
            return self.separator.join(value)
        return value

    def value_to_string(self, obj):
        value = self.value_from_object(obj)
        return self.get_prep_value(value)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs['separator'] = self.separator
        return name, path, args, kwargs


class HexColorField(models.CharField):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('max_length', 7)
        super().__init__(*args, **kwargs)

    def formfield(self, **kwargs):
        """Use custom widget for both regular and admin forms"""
        widget = kwargs.get("widget", ColorisWidget)

        if widget == admin_widgets.AdminTextInputWidget:
            widget = dashub_widgets.AdminColorisWidget()

        kwargs["widget"] = widget
        return super().formfield(**kwargs)






from typing import Optional, Generator, Any

from django.contrib.admin import ModelAdmin
from django.contrib.admin.views.main import ChangeList
from django.core.validators import EMPTY_VALUES
from django.db.models import QuerySet
from django.http import HttpRequest
from django.utils.translation import gettext_lazy as _
from django.db.models.fields import BLANK_CHOICE_DASH, Field
from dashub.contrib.forms import DropdownForm, AutocompleteDropdownForm


class ValueMixin:
    def value(self) -> Optional[str]:
        return (
            self.lookup_val[0]
            if self.lookup_val not in EMPTY_VALUES
               and isinstance(self.lookup_val, list)
               and len(self.lookup_val) > 0
            else self.lookup_val
        )


class MultiValueMixin:
    def value(self) -> Optional[list[str]]:
        return (
            self.lookup_val
            if self.lookup_val not in EMPTY_VALUES
            and isinstance(self.lookup_val, list)
            and len(self.lookup_val) > 0
            else self.lookup_val
        )


class ChoicesMixin:
    template = "dashub/filters/filters_field.html"

    def choices(self, changelist: ChangeList) -> Generator[dict[str, Any], None, None]:
        add_facets = getattr(changelist, "add_facets", False)
        facet_counts = self.get_facet_queryset(changelist) if add_facets else None
        choices = [self.all_option] if self.all_option else []

        for i, choice in enumerate(self.field.flatchoices):
            if add_facets:
                count = facet_counts[f"{i}__c"]
                choice = (choice[0], f"{choice[1]} ({count})")

            choices.append(choice)

        yield {
            "form": self.form_class(
                label=_(" By %(filter_title)s ") % {"filter_title": self.title},
                name=self.lookup_kwarg,
                choices=choices,
                data={self.lookup_kwarg: self.value()},
            ),
        }


class DropdownMixin:
    template = "dashub/filters/filters_field.html"
    form_class = DropdownForm
    all_option = ["", _("All")]

    def queryset(self, request: HttpRequest, queryset: QuerySet) -> QuerySet:
        if self.value() not in EMPTY_VALUES:
            return super().queryset(request, queryset)

        return queryset



class AutocompleteMixin:
    def has_output(self) -> bool:
        return True

    def field_choices(
        self, field: Field, request: HttpRequest, model_admin: ModelAdmin
    ) -> list[tuple[str, list[tuple[str, str]]]]:
        return [
            ("", BLANK_CHOICE_DASH),
        ]

    def choices(
        self, changelist: ChangeList
    ) -> Generator[dict[str, AutocompleteDropdownForm], None, None]:
        yield {
            "form": self.form_class(
                request=self.request,
                label=_(" By %(filter_title)s ") % {"filter_title": self.title},
                name=self.lookup_kwarg,
                choices=(),
                field=self.field,
                model_admin=self.model_admin,
                data={self.lookup_kwarg: self.value()},
                multiple=self.multiple if hasattr(self, "multiple") else False,
            ),
        }





from django import template
from django.conf import settings

register = template.Library()

@register.filter
def is_crispy_form(form):
    """
    Check if the form is a crispy form.
    """
    try:
        from crispy_forms.helper import FormHelper
    except ImportError:
        return False

    if not "crispy_forms" in settings.INSTALLED_APPS:
        return False

    return hasattr(form, 'helper') and isinstance(form.helper, FormHelper)

@register.simple_tag(takes_context=True)
def safe_crispy(context, form):
    """
    Render form with crispy if crispy is installed and form is crispy,
    else just render the form as HTML.
    """
    try:
        from crispy_forms.utils import render_crispy_form
        return render_crispy_form(form, None, context)
    except ImportError:
        pass
    return str(form)



from rest_framework import serializers

class TagInputField(serializers.Field):
    def to_representation(self, value):
        if isinstance(value, str):
            return value.split(":::") if value else []
        return value

    def to_internal_value(self, data):
        if isinstance(data, list):
            return ":::" .join(data)
        raise serializers.ValidationError("Expected a list of strings.")



import copy
import importlib
import itertools
import json
import logging
import urllib.parse
from datetime import timedelta
from typing import Any, Callable, Dict, List, Optional, Union

from django.apps import apps
from django.conf import settings
from django.contrib.admin import ListFilter
from django.contrib.admin.helpers import AdminForm, Fieldset, InlineAdminFormSet
from django.contrib.admin.models import LogEntry
from django.contrib.admin.sites import all_sites
from django.contrib.admin.views.main import PAGE_VAR, ChangeList, ALL_VAR
from django.contrib.auth import get_user_model
from django.contrib.auth.context_processors import PermWrapper
from django.contrib.auth.models import AbstractUser
from django.core.handlers.wsgi import WSGIRequest
from django.db.models.base import ModelBase
from django.http import HttpRequest
from django.template import Context, Library
from django.template.defaultfilters import capfirst
from django.template.loader import get_template
from django.templatetags.static import static
from django.urls import reverse, NoReverseMatch
from django.utils import timezone
from django.utils.html import escape, format_html
from django.utils.safestring import SafeText, mark_safe
from django.utils.text import get_text_list, slugify
from django.utils.translation import gettext
from django.forms import widgets, JSONField

from .. import version
from ..settings import CHANGEFORM_TEMPLATES, get_resolved_settings
from ..utils import (
    get_admin_url,
    get_filter_id,
    get_installed_apps,
    has_fieldsets_check,
    make_menu,
    order_with_respect_to,
)

User = get_user_model()
register = Library()
logger = logging.getLogger(__name__)


def get_model_info(model_path: str):
    """
    Fetch model verbose name and admin list URL dynamically.

    :param model_path: Model path in "app_label.ModelName" format.
    :return: Dictionary with 'name' and 'url'.
    """
    try:
        app_label, model_name = model_path.split(".")
        model = apps.get_model(app_label, model_name)

        verbose_name = model._meta.verbose_name_plural or model._meta.verbose_name
        model_admin_url = reverse(f"admin:{app_label}_{model_name.lower()}_changelist")
        return {"name": str(verbose_name), "url": model_admin_url}
    except Exception as e:
        return {"name": model_path, "url": "#"}


@register.simple_tag(takes_context=True)
def get_side_menu(context: Context, using: str = "available_apps") -> List[Dict[str, Any]]:
    """
    Generate the side menu with sorting, custom submenus, and additional custom links.
    """
    user = context.get("user")
    if not user:
        return []

    options = get_resolved_settings()
    default_orders = options.get("default_orders", {})
    installed_apps = get_installed_apps()
    available_apps = copy.deepcopy(context.get(using, []))
    request = context.get("request")

    if not available_apps:
        from django.contrib import admin
        available_apps = admin.site.get_app_list(request)

    # Add custom apps if not already available
    for app_label, links in options.get("custom_links", {}).items():
        if app_label.lower() not in installed_apps:
            available_apps.append({
                "name": app_label,
                "app_label": app_label,
                "app_url": "#",
                "has_module_perms": True,
                "models": [],
                "order": links["order"] if "order" in links else default_orders.get(app_label, 0),
            })

    custom_links = {
        app_name: make_menu(user, links, options, app_name, allow_appmenus=False)
        for app_name, links in options.get("custom_links", {}).items()
    }

    submenus_models = set(options.get("submenus_models", []))
    hidden_apps = set(options.get("hide_apps", []))
    hidden_models = set(options.get("hide_models", []))
    order_menus = options.get("default_orders", [])

    menu = []

    for app in available_apps:
        app_label = app["app_label"]
        if app_label in hidden_apps:
            continue

        app["icon"] = options["icons"].get(app_label, options["default_icon_parents"])
        app["order"] = order_menus[app_label] if app_label in order_menus else default_orders.get(app_label, 1000)

        menu_items = []

        app_custom_links = custom_links.get(app_label, [])
        for model in app.get("models", []):
            model_object_name = str(model['object_name']).lower()
            model_str = f"{app_label}.{model_object_name}".lower()
            if model_str in hidden_models:
                continue

            model["url"] = model["admin_url"]
            model["count"] = model.get("model", {}).objects.count() if model.get("model") else 0
            model["model_str"] = model_str
            model["icon"] = options["icons"].get(model_str, options["default_icon_children"])
            model["order"] = model.get("order", default_orders.get(model_str, 1000))

            finded_menus = {}
            if app_custom_links:
                finded_menus = next((item for item in app_custom_links if item['object_name'] == model_str), None)

            if model_str in submenus_models:
                submenu = [{"name": "Add New", "url": model["add_url"], "order": 0},
                           {"name": model["name"], "url": model["url"], "order": 0}]
            else:
                submenu = []
            if finded_menus:
                custom_links_submenus = finded_menus.get("submenu", [])
                submenu.extend(custom_links_submenus)
                app_custom_links = [link for link in app_custom_links if link['name'] != model_object_name]
            model["submenu"] = sorted(submenu, key=lambda x: x.get("order", 1000), reverse=True)
            menu_items.append(model)
        if app_custom_links:
            for new_link in app_custom_links:
                if not any(
                        (existing_link.get('url') == new_link.get('url')) or
                        (existing_link.get('model_str') == new_link.get('object_name'))
                        for existing_link in menu_items
                ):
                    menu_items.append(new_link)

        if menu_items:
            app["models"] = sorted(menu_items, key=lambda x: x.get("order", 1000))
            menu.append(app)

    return order_menus_with_order(menu, order_menus)


def order_menus_with_order(menu: List[Dict[str, Any]], order_menus: List[str]) -> List[Dict[str, Any]]:
    """
    Sorts menu items based on `order_menus` list. Items not in the list will retain their original order.
    This function ensures sorting in the following hierarchy: submenus → models → apps.
    """
    order_map = {label.lower(): index for index, label in enumerate(order_menus)}

    def get_order_index(item: Dict[str, Any]) -> int:
        """Returns the order index of an app."""
        return order_map.get(item["app_label"].lower(), item.get("order", 1000))

    def sort_models(models: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """Sort models inside apps based on 'order' key."""
        for model in models:
            if "submenu" in model:
                model["submenu"].sort(key=lambda x: x.get("order", 1000), reverse=False)
        return sorted(models, key=lambda x: x.get("order", 1000), reverse=False)

    for app in menu:
        if "models" in app:
            app["models"] = sort_models(app["models"])

    return sorted(menu, key=get_order_index, reverse=False)


@register.simple_tag
def get_top_menu(user: AbstractUser, admin_site: str = "admin") -> List[Dict]:
    """
    Produce the menu for the top nav bar
    """
    options = get_resolved_settings()
    return make_menu(user, options.get("topmenu_links", []), options, None, allow_appmenus=True, admin_site=admin_site)


@register.simple_tag
def get_user_menu(user: AbstractUser, admin_site: str = "admin") -> List[Dict]:
    """
    Produce the menu for the user dropdown
    """
    options = get_resolved_settings()
    return make_menu(
        user,
        options.get("usermenu_links", []),
        options,
        None,
        allow_appmenus=False,
        admin_site=admin_site,
    )


@register.simple_tag
def get_dashub_settings(request: WSGIRequest) -> Dict:
    """
    Get Practet Dashboard settings, update any defaults from the request, and return
    """
    settings = get_resolved_settings(request)

    admin_site = {x.name: x for x in all_sites}.get("admin", {})
    if not settings["site_title"]:
        settings["site_title"] = getattr(admin_site, "site_title", None)

    if not settings["site_header"]:
        settings["site_header"] = getattr(admin_site, "site_header", None)

    if not settings["site_brand"]:
        settings["site_brand"] = getattr(admin_site, "site_header", None)

    return settings


@register.simple_tag
def get_dashub_version() -> str:
    """
    Get the version for this package
    """
    return version


@register.simple_tag
def get_user_avatar(user: AbstractUser) -> str:
    """
    For the given user, try to get the avatar image, which can be one of:

        - ImageField on the user model
        - URLField/Charfield on the model
        - A callable that receives the user instance e.g lambda u: u.profile.image.url
    """
    no_avatar = static("assets/img/user.svg")
    options = get_resolved_settings()
    avatar_field_name: Optional[Union[str, Callable]] = options.get("user_avatar")

    if not avatar_field_name:
        return no_avatar

    if callable(avatar_field_name):
        return avatar_field_name(user)

    # If we find the property directly on the user model (imagefield or URLfield)
    avatar_field = getattr(user, avatar_field_name, None)
    if avatar_field is not None:
        if not avatar_field:
            return no_avatar
        if isinstance(avatar_field, str):
            return avatar_field
        elif hasattr(avatar_field, "url"):
            return avatar_field.url
        elif callable(avatar_field):
            return avatar_field()

    logger.warning("Avatar field must be an ImageField/URLField on the user model, or a callable")

    return no_avatar


@register.simple_tag
def dashub_paginator_number(change_list: ChangeList, i: int) -> SafeText:
    """
    Generate an individual page index link in a paginated list.
    """
    html_str = ""
    start = i == 1
    end = i == change_list.paginator.num_pages
    spacer = i in (".", "…")
    current_page = i == change_list.page_num

    if start:
        link = change_list.get_query_string({PAGE_VAR: change_list.page_num - 1}) if change_list.page_num > 1 else "#"
        html_str += """
        <li class="page-item previous {disabled}">
            <a class="page-link" href="{link}" data-dt-idx="0" tabindex="0">Previous</a>
        </li>
        """.format(link=link, disabled="disabled" if link == "#" else "")

    if current_page:
        html_str += """
        <li class="page-item active">
            <a class="page-link" href="javascript:void(0);" data-dt-idx="3" tabindex="0">{num}</a>
        </li>
        """.format(num=i)
    elif spacer:
        html_str += """
        <li class="page-item">
            <a class="page-link" href="javascript:void(0);" data-dt-idx="3" tabindex="0">… </a>
        </li>
        """
    else:
        query_string = change_list.get_query_string({PAGE_VAR: i})
        end = "end" if end else ""
        html_str += """
            <li class="page-item">
            <a href="{query_string}" class="page-link {end}" data-dt-idx="3" tabindex="0">{num}</a>
            </li>
        """.format(num=i, query_string=query_string, end=end)

    if end:
        link = change_list.get_query_string({PAGE_VAR: change_list.page_num + 1}) if change_list.page_num < i else "#"
        html_str += """
        <li class="page-item next {disabled}">
            <a class="page-link" href="{link}" data-dt-idx="7" tabindex="0">Next</a>
        </li>
        """.format(link=link, disabled="disabled" if link == "#" else "")

    return format_html(html_str)


@register.simple_tag
def admin_extra_filters(cl: ChangeList) -> Dict:
    """
    Return the dict of used filters which is not included in list_filters form
    """
    used_parameters = list(itertools.chain(*(s.used_parameters.keys() for s in cl.filter_specs)))
    return {k: v for k, v in cl.params.items() if k not in used_parameters}


@register.simple_tag
def dashub_list_filter(cl: ChangeList, spec: ListFilter):
    """
    Render out our list filter in a dropdown friendly format, for use by filter.html, see original implementation here

    django.contrib.admin.templatetags.admin_list.admin_list_filter

    """
    tpl = get_template(spec.template)
    choices = list(spec.choices(cl))
    field_key = get_filter_id(spec)
    field_name = ""

    for choice in choices:
        qs = choice.get("query_string")
        if not qs:
            continue

        matches = {}
        query_parts = urllib.parse.parse_qs(qs[1:])
        if f"{field_key}__exact" in query_parts:
            matches[f"{field_key}__exact"] = query_parts[f"{field_key}__exact"][0]
        elif field_key in query_parts:
            matches[field_key] = query_parts[field_key][0]
        else:
            for key in query_parts:
                if key.startswith(field_key + "__") or f"__{field_key}__" in key:
                    matches[key] = query_parts[key][0]

        # Iterate matches, use original as actual values, additional for hidden
        for i, (key, value) in enumerate(matches.items()):
            if i == 0:
                choice["name"] = key
                choice["value"] = value
                field_name = key

    return tpl.render({
        "field_name": field_name,
        "title": spec.title,
        "choices": choices,
        "spec": spec
    })


@register.simple_tag
def jazzy_admin_url(value: Union[str, ModelBase], admin_site: str = "admin") -> str:
    """
    Get the admin url for a given object
    """
    return get_admin_url(value, admin_site=admin_site)


@register.filter
def has_dashub_setting(settings: Dict[str, Any], key: str) -> bool:
    return key in settings and settings[key] is not None


@register.filter
def has_fieldsets(adminform: AdminForm) -> bool:
    """
    Do we have fieldsets
    """
    return has_fieldsets_check(adminform)


@register.simple_tag
def get_sections(
        admin_form: AdminForm, inline_admin_formsets: List[InlineAdminFormSet]
) -> List[Union[Fieldset, InlineAdminFormSet]]:
    """
    Get and sort all of the sections that need rendering out in a change form
    """
    fieldsets = list(admin_form)

    # Make inlines behave like formsets
    for fieldset in inline_admin_formsets:
        fieldset.name = capfirst(fieldset.opts.verbose_name_plural)
        fieldset.is_inline = True
        fieldsets.append(fieldset)

    if hasattr(admin_form.model_admin, "dashub_section_order"):
        fieldsets = order_with_respect_to(
            fieldsets, admin_form.model_admin.dashub_section_order, getter=lambda x: x.name
        )

    return fieldsets


@register.filter
def debug(value: Any) -> Any:
    """
    Add in a breakpoint() here and use filter in templates for debugging ;)
    """
    return type(value)


@register.filter
def as_json(value: Union[List, Dict]) -> str:
    """
    Take the given item and dump it out as JSON
    """
    return json.dumps(value)


@register.simple_tag
def get_changeform_template(adminform: AdminForm) -> str:
    """
    Go get the correct change form template based on the modeladmin being used,
    the default template, or the overridden one for this modeladmin
    """
    options = get_resolved_settings()
    has_fieldsets = has_fieldsets_check(adminform)
    inlines = adminform.model_admin.inlines
    has_inlines = inlines and len(inlines) > 0
    model = adminform.model_admin.model
    model_name = "{}.{}".format(model._meta.app_label, model._meta.model_name).lower()

    changeform_format = options.get("changeform_format", "")
    if model_name in options.get("changeform_format_overrides", {}):
        changeform_format = options["changeform_format_overrides"][model_name]

    if not has_fieldsets and not has_inlines:
        return CHANGEFORM_TEMPLATES["single"]

    if not changeform_format or changeform_format not in CHANGEFORM_TEMPLATES.keys():
        return CHANGEFORM_TEMPLATES["horizontal_tabs"]

    return CHANGEFORM_TEMPLATES[changeform_format]


@register.simple_tag
def get_changeform_template_class(adminform: AdminForm) -> str:
    """
    Go get the correct change form template based on the modeladmin being used,
    the default template, or the overridden one for this modeladmin
    """
    options = get_resolved_settings()
    has_fieldsets = has_fieldsets_check(adminform)
    inlines = adminform.model_admin.inlines
    has_inlines = inlines and len(inlines) > 0
    model = adminform.model_admin.model
    model_name = "{}.{}".format(model._meta.app_label, model._meta.model_name).lower()

    changeform_format = options.get("changeform_format", "")
    if model_name in options.get("changeform_format_overrides", {}):
        changeform_format = options["changeform_format_overrides"][model_name]

    if not has_fieldsets and not has_inlines:
        return "single"

    if not changeform_format or changeform_format not in CHANGEFORM_TEMPLATES.keys():
        return "horizontal_tabs"

    return changeform_format


@register.filter
def can_view_self(perms: PermWrapper) -> bool:
    """
    Determines whether a user has sufficient permissions to view its own profile
    """
    view_perm = "view_{}".format(User._meta.model_name)
    return perms[User._meta.app_label][view_perm]


@register.simple_tag
def header_class(header: Dict, forloop: Dict) -> str:
    """
    Adds CSS classes to header HTML element depending on its attributes
    """
    classes = []
    sorted, asc, desc = (
        header.get("sorted"),
        header.get("ascending"),
        header.get("descending"),
    )

    is_checkbox_column_conditions = (
        forloop["counter0"] == 0,
        header.get("class_attrib") == ' class="action-checkbox-column"',
    )
    if all(is_checkbox_column_conditions):
        classes.append("djn-checkbox-select-all")

    if not header["sortable"]:
        return " ".join(classes)

    if sorted and asc:
        classes.append("sorting_asc")
    elif sorted and desc:
        classes.append("sorting_desc")
    else:
        classes.append("sorting")

    return " ".join(classes)


@register.filter
def app_is_installed(app: str) -> bool:
    """
    Checks if an app has been installed under INSTALLED_APPS on the project settings
    """
    return app in settings.INSTALLED_APPS


@register.simple_tag
def action_message_to_list(action: LogEntry) -> List[Dict]:  # noqa: C901
    """
    Retrieves a formatted list with all actions taken by a user given a log entry object
    """
    messages = []

    def added(x: str) -> Dict:
        return {
            "msg": x,
            "icon": "plus-circle",
            "colour": "success",
        }

    def changed(x: str) -> Dict:
        return {
            "msg": x,
            "icon": "edit",
            "colour": "blue",
        }

    def deleted(x: str) -> Dict:
        return {
            "msg": x,
            "icon": "trash",
            "colour": "danger",
        }

    if action.change_message and action.change_message[0] == "[":
        try:
            change_message = json.loads(action.change_message)
        except json.JSONDecodeError:
            return [action.change_message]

        for sub_message in change_message:
            if "added" in sub_message:
                if sub_message["added"]:
                    sub_message["added"]["name"] = gettext(sub_message["added"]["name"])
                    messages.append(added(gettext("Added {name} “{object}”.").format(**sub_message["added"])))
                else:
                    messages.append(added(gettext("Added.")))

            elif "changed" in sub_message:
                sub_message["changed"]["fields"] = get_text_list(
                    [gettext(field_name) for field_name in sub_message["changed"]["fields"]],
                    gettext("and"),
                )
                if "name" in sub_message["changed"]:
                    sub_message["changed"]["name"] = gettext(sub_message["changed"]["name"])
                    messages.append(changed(gettext("Changed {fields}.").format(**sub_message["changed"])))
                else:
                    messages.append(changed(gettext("Changed {fields}.").format(**sub_message["changed"])))

            elif "deleted" in sub_message:
                sub_message["deleted"]["name"] = gettext(sub_message["deleted"]["name"])
                messages.append(deleted(gettext("Deleted “{object}”.").format(**sub_message["deleted"])))

    return messages if len(messages) else [changed(gettext(action.change_message))]


@register.filter
def style_bold_first_word(message: str) -> str | Any:
    """
    Wraps first word in a message with <strong> HTML element
    """
    message_words = escape(message).split()

    if not len(message_words):
        return ""

    message_words[0] = "<strong>{}</strong>".format(message_words[0])

    message = " ".join(list(message_words))

    return mark_safe(message)


@register.filter
def unicode_slugify(message: str) -> str:
    return slugify(message, allow_unicode=True)


@register.filter
def render_form_field_class(field, css_class=None):
    widget = field.field.widget
    existing_classes = widget.attrs.get("class", "")
    class_list = existing_classes.split()

    if isinstance(field.field, JSONField):
        default_class = "form-control jsoneditor"
    elif isinstance(widget, (
            widgets.Select,
            widgets.SelectMultiple
    )):
        default_class = "form-select"
    elif isinstance(widget, (
            widgets.CheckboxInput,
            widgets.CheckboxSelectMultiple,
            widgets.RadioSelect
    )):
        default_class = None
    else:
        default_class = "form-control"

    if css_class is None:
        css_class = default_class

    if css_class and css_class not in class_list:
        class_list.append(css_class)

    if field.errors and "is-invalid" not in class_list:
        class_list.append("is-invalid")

    widget.attrs["class"] = " ".join(class_list)
    return field


@register.filter
def is_path_image(path: str) -> bool:
    """
    Check if the given path is an image
    """
    return path.lower().endswith((".png", ".jpg", ".jpeg", ".gif", ".avif", ".svg", ".webp"))


@register.filter
def class_name(value: Any) -> str:
    return value.__class__.__name__


@register.filter
def pagination_show_all_url(cl) -> bool:
    need_show_all_link = cl.can_show_all and not cl.show_all and cl.multi_page
    return need_show_all_link and cl.get_query_string({ALL_VAR: ""})


@register.inclusion_tag(
    "dashub/templatetags/preserve_changelist_filters.html",
    takes_context=True,
    name="preserve_filters",
)
def preserve_changelist_filters(context: Context) -> dict[str, dict[str, str]]:
    """
    Generate hidden input fields to preserve filters for POST forms.
    """
    request: Optional[HttpRequest] = context.get("request")
    changelist: Optional[ChangeList] = context.get("cl")

    if not request or not changelist:
        return {"params": {}}

    used_params: set[str] = {
        param for spec in changelist.filter_specs for param in spec.used_parameters
    }
    preserved_params: dict[str, str] = {
        param: value for param, value in request.GET.items() if param not in used_params
    }

    return {"params": preserved_params}


@register.filter
def is_crispy_form(form: Any) -> bool:
    """
    Check if the form has a crispy FormHelper attached.
    """
    try:
        from crispy_forms.helper import FormHelper
    except ImportError:
        return False

    if not "crispy_forms" in settings.INSTALLED_APPS:
        return False

    return hasattr(form, 'helper') and isinstance(form.helper, FormHelper)



@register.simple_tag
def crispy_available():
    try:
        if not "crispy_forms" in settings.INSTALLED_APPS:
            return False

        importlib.import_module("crispy_forms")
        return True
    except ImportError:
        return False


@register.simple_tag(takes_context=True)
def get_theme(context):
    request = context['request']
    cookie_theme_choice = request.COOKIES.get('dashub_theme')
    cookie_resolved_theme = request.COOKIES.get('dashub_theme_resolved')
    dashub_settings = get_dashub_settings(request)

    theme_from_settings = getattr(dashub_settings, "theme") if hasattr(dashub_settings, "theme") else "system"
    theme_choice = cookie_theme_choice or theme_from_settings or 'system'

    if theme_choice == 'system':
        if cookie_resolved_theme:
            return cookie_resolved_theme
        else:
            return 'light'
    else:
        return theme_choice


@register.simple_tag
def safe_admin_changelist_url(opts):
    try:
        return reverse(f'admin:{opts.app_label}_{opts.model_name}_changelist')
    except NoReverseMatch:
        return None



















from setuptools import setup, find_packages

setup(
    name="django_dashub",
    version="3.0.7",
    author="Suresh Chand",
    author_email="scthakuri12a@gmail.com",
    description="A modern Django admin dashboard with enhanced customization options, inspired by Jazzmin but featuring a fresh theme and additional functionality.",
    packages=find_packages(),
    license="BSD",
    install_requires=[
        "django",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Framework :: Django",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: BSD License"
    ],
    include_package_data=True,
    package_data={
        "django_dashub": [
            "templates/*",
            "static/*",
            "contrib/*",
            "templatetags/*",
            "AUTHORS",
            "LICENSE",
            "README.md"
        ],
    },
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    project_urls={
        "Bug Reports": "https://github.com/scthakuri/django-dashub/issues",
        "Source": "https://github.com/scthakuri/django-dashub",
    },
)



from typing import Optional, Any

from django import forms
from django.contrib.admin.widgets import AdminRadioSelect, AdminSplitDateTime, AdminDateWidget, AdminTimeWidget
from django.forms.widgets import Select, SelectMultiple, CheckboxSelectMultiple, MultiWidget, TextInput
from django.contrib.admin import widgets as admin_widgets, VERTICAL
from django.utils.translation import gettext_lazy as _


class DashubSelect(Select):
    template_name = "dashub/widgets/select.html"

    def build_attrs(self, base_attrs, extra_attrs=None):
        return {**base_attrs, **(extra_attrs or {})}

    @property
    def media(self):
        return forms.Media(
            css={"all": ("https://cdn.jsdelivr.net/npm/select2@4.1.0-rc.0/dist/css/select2.min.css",)},
            js=("https://cdn.jsdelivr.net/npm/select2@4.1.0-rc.0/dist/js/select2.min.js",),
        )


class DashubSelectMultiple(SelectMultiple):
    template_name = "dashub/widgets/select.html"

    def build_attrs(self, base_attrs, extra_attrs=None):
        extra_attrs["multiple"] = "multiple"
        return {**base_attrs, **(extra_attrs or {})}

    @property
    def media(self):
        return forms.Media(
            css={"all": ("https://cdn.jsdelivr.net/npm/select2@4.1.0-rc.0/dist/css/select2.min.css",)},
            js=("https://cdn.jsdelivr.net/npm/select2@4.1.0-rc.0/dist/js/select2.min.js",),
        )


class TagInputWidget(forms.Textarea):
    template_name = "dashub/widgets/tag_input.html"

    def __init__(self, attrs=None, separator=":::"):
        self.separator = separator
        final_attrs = {"data-separator": separator}
        if attrs:
            final_attrs.update(attrs)
        super().__init__(attrs=final_attrs)

    class Media:
        js = (
            "https://cdn.jsdelivr.net/npm/@yaireo/tagify",
            "https://cdn.jsdelivr.net/npm/@yaireo/tagify/dist/tagify.polyfills.min.js"
        )
        css = {"all": ("https://cdn.jsdelivr.net/npm/@yaireo/tagify/dist/tagify.css",)}

    def format_value(self, value):
        if value is not None:
            if isinstance(value, list):
                return self.separator.join(value)
            elif isinstance(value, str):
                return value.strip()
        return ""


class AdminTagInputWidget(TagInputWidget, admin_widgets.AdminTextareaWidget):
    pass


class DashubAdminRadioSelectWidget(AdminRadioSelect):
    template_name = "dashub/widgets/radio.html"
    option_template_name = "dashub/widgets/radio_option.html"
    RADIO_CLASSES = ["form-check-input"]

    def __init__(self, radio_style: Optional[int] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if radio_style is None:
            radio_style = VERTICAL
        self.radio_style = radio_style
        self.attrs["class"] = " ".join([*self.RADIO_CLASSES, self.attrs.get("class", "")])

    def get_context(self, *args, **kwargs) -> dict[str, Any]:
        context = super().get_context(*args, **kwargs)
        context.update({"radio_style": self.radio_style})
        return context


class DashubAdminCheckboxSelectMultiple(CheckboxSelectMultiple):
    template_name = "dashub/widgets/radio.html"
    option_template_name = "dashub/widgets/radio_option.html"
    CHECKBOX_CLASSES = ["form-check-input"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attrs["class"] = " ".join([*self.CHECKBOX_CLASSES, self.attrs.get("class", "")])


class DashubAdminDateWidget(AdminDateWidget):
    template_name = "dashub/widgets/date.html"
    DATETIME_CLASSES = ["form-control"]

    def __init__(
        self, attrs: Optional[dict[str, Any]] = None, format: Optional[str] = None
    ) -> None:
        attrs = {
            **(attrs or {}),
            "class": " ".join(
                [
                    "vDateField",
                    *self.DATETIME_CLASSES,
                    attrs.get("class", "") if attrs else "",
                ]
            ),
            "size": "10",
        }
        super().__init__(attrs=attrs, format=format)


class DashubAdminTimeWidget(AdminTimeWidget):
    template_name = "dashub/widgets/time.html"
    DATETIME_CLASSES = ["form-control"]

    def __init__(
        self, attrs: Optional[dict[str, Any]] = None, format: Optional[str] = None
    ) -> None:
        attrs = {
            **(attrs or {}),
            "class": " ".join(
                [
                    "vTimeField",
                    *self.DATETIME_CLASSES,
                    attrs.get("class", "") if attrs else "",
                ]
            ),
            "size": "8",
        }
        super().__init__(attrs=attrs, format=format)


class DashubAdminSplitDateTimeVerticalWidget(AdminSplitDateTime):
    template_name = "dashub/widgets/split_datetime_vertical.html"

    def __init__(
        self,
        attrs: Optional[dict[str, Any]] = None,
        date_attrs: Optional[dict[str, Any]] = None,
        time_attrs: Optional[dict[str, Any]] = None,
        date_label: Optional[str] = None,
        time_label: Optional[str] = None,
    ) -> None:
        self.date_label = date_label
        self.time_label = time_label

        widgets = [
            DashubAdminDateWidget(attrs=date_attrs),
            DashubAdminTimeWidget(attrs=time_attrs),
        ]
        MultiWidget.__init__(self, widgets, attrs)

    def get_context(
        self, name: str, value: Any, attrs: Optional[dict[str, Any]]
    ) -> dict[str, Any]:
        context = super().get_context(name, value, attrs)

        if self.date_label is not None:
            context["date_label"] = self.date_label
        else:
            context["date_label"] = _("Date")

        if self.time_label is not None:
            context["time_label"] = self.time_label
        else:
            context["time_label"] = _("Time")

        return context


class ColorisWidget(TextInput):
    def build_attrs(self, base_attrs, extra_attrs=None):
        extra_attrs["color-picker"] = "color-picker"
        return {**base_attrs, **(extra_attrs or {})}

    @property
    def media(self):
        return forms.Media(
            css={"all": ("https://cdn.jsdelivr.net/gh/mdbassit/Coloris@latest/dist/coloris.min.css",)},
            js=("https://cdn.jsdelivr.net/gh/mdbassit/Coloris@latest/dist/coloris.min.js",),
        )


class AdminColorisWidget(ColorisWidget, admin_widgets.AdminTextInputWidget):
    pass







from django.contrib.admin.widgets import AutocompleteSelect, AutocompleteSelectMultiple
from django.forms import MultipleChoiceField, ChoiceField, CheckboxSelectMultiple, RadioSelect, ModelMultipleChoiceField
from django import forms
from django.http import HttpRequest
from django.utils.translation import gettext_lazy as _
from dashub.widgets import DashubAdminCheckboxSelectMultiple, DashubAdminRadioSelectWidget, DashubSelect, \
    DashubSelectMultiple, DashubAdminSplitDateTimeVerticalWidget


class CheckboxForm(forms.Form):
    field = MultipleChoiceField
    widget = DashubAdminCheckboxSelectMultiple

    def __init__(
        self,
        name: str,
        label: str,
        choices: tuple,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)

        self.fields[name] = self.field(
            label=label,
            required=False,
            choices=choices,
            widget=self.widget,
        )


class RadioForm(CheckboxForm):
    field = ChoiceField
    widget = DashubAdminRadioSelectWidget


class DropdownForm(forms.Form):
    widget = DashubSelect()
    field = ChoiceField

    def __init__(
        self,
        name: str,
        label: str,
        choices: tuple,
        multiple: bool = False,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)

        if multiple:
            self.widget = DashubSelectMultiple()
            self.field = MultipleChoiceField

        self.fields[name] = self.field(
            label=label,
            required=False,
            choices=choices,
            widget=self.widget,
        )



class AutocompleteDropdownForm(forms.Form):
    field = forms.ModelChoiceField
    widget = AutocompleteSelect

    def __init__(
        self,
        request: HttpRequest,
        name: str,
        label: str,
        choices: tuple,
        field,
        model_admin,
        multiple: bool = False,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)

        if multiple:
            self.field = ModelMultipleChoiceField
            self.widget = AutocompleteSelectMultiple

        self.fields[name] = self.field(
            label=label,
            required=False,
            queryset=field.remote_field.model.objects,
            widget=self.widget(field, model_admin.admin_site),
        )


class RangeDateForm(forms.Form):
    INPUT_CLASSES = ["form-control"]

    def __init__(self, name: str, *args, **kwargs) -> None:
        self.name = name
        super().__init__(*args, **kwargs)

        self.fields[self.name + "_from"] = forms.DateField(
            label="",
            required=False,
            widget=forms.DateInput(
                attrs={
                    "placeholder": _("From"),
                    "class": "vDateField vCustomDateField " + " ".join(self.INPUT_CLASSES),
                }
            ),
        )
        self.fields[self.name + "_to"] = forms.DateField(
            label="",
            required=False,
            widget=forms.DateInput(
                attrs={
                    "placeholder": _("To"),
                    "class": "vDateField vCustomDateField " + " ".join(self.INPUT_CLASSES),
                }
            ),
        )


class RangeDateTimeForm(forms.Form):
    INPUT_CLASSES = ["form-control"]
    def __init__(self, name: str, *args, **kwargs) -> None:
        self.name = name
        super().__init__(*args, **kwargs)

        self.fields[self.name + "_from"] = forms.SplitDateTimeField(
            label="",
            required=False,
            widget=DashubAdminSplitDateTimeVerticalWidget(
                date_label="",
                date_attrs={
                    "placeholder": _("Date from"),
                    "class": "vCustomDateField " + " ".join(self.INPUT_CLASSES),
                },
                time_label="",
                time_attrs={
                    "placeholder": _("Time"),
                    "class": "vCustomTimeField " + " ".join(self.INPUT_CLASSES),
                },
            ),
        )
        self.fields[self.name + "_to"] = forms.SplitDateTimeField(
            label="",
            required=False,
            widget=DashubAdminSplitDateTimeVerticalWidget(
                date_label="",
                date_attrs={
                    "placeholder": _("Date to"),
                    "class": "vCustomDateField " + " ".join(self.INPUT_CLASSES),
                },
                time_label="",
                time_attrs={
                    "placeholder": _("Time"),
                    "class": "vCustomTimeField " + " ".join(self.INPUT_CLASSES),
                },
            ),
        )





import datetime
import importlib
import inspect
import logging
from typing import Any, Callable, Dict, List, Set, Union, Optional
from urllib.parse import urlencode

from django.apps import apps
from django.conf import settings
from django.contrib.admin import ListFilter
from django.contrib.admin.helpers import AdminForm
from django.contrib.auth.models import AbstractUser
from django.core.handlers.wsgi import WSGIRequest
from django.db.models.base import Model, ModelBase
from django.db.models.options import Options
from django.utils import timezone
from django.utils.text import slugify
from django.utils.translation import gettext

from .compat import NoReverseMatch, reverse

logger = logging.getLogger(__name__)


def order_with_respect_to(original: List, reference: List, getter: Callable = lambda x: x) -> List:
    """
    Order a list based on the location of items in the reference list, optionally, use a getter to pull values out of
    the first list
    """
    ranking = []
    max_num = len(original)
    for item in original:
        try:
            pos = reference.index(getter(item))
        except ValueError:
            pos = max_num

        ranking.append(pos)

    return [y for x, y in sorted(zip(ranking, original), key=lambda x: x[0])]


def order_menus_with_order(original: List[Dict[str, Any]], order_list: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    for item in original:
        item["order"] = item.get("item", 0)
        app_label = item.get("app_label") or item.get("app")
        order_entry = next((entry for entry in order_list if entry["app"] == app_label), None)
        if order_entry:
            item["order"] = order_entry["order"]

        if "models" in item and isinstance(item["models"], list):
            model_order_list = order_entry.get("models", []) if order_entry else []

            for model in item["models"]:
                model["order"] = model.get("order", 0)
                model_entry = next((m for m in model_order_list if m["model"] == model.get("object_name")), None)
                if model_entry:
                    model["order"] = model_entry["order"]

                # Also order submenus if they exists
                if "submenus" in model and isinstance(model["submenus"], list):
                    model["submenus"] = [submenu.get("order", 0) for submenu in model["submenus"]]
                    model["submenus"] = model["submenus"].sort(key=lambda x: x["order"], reverse=True)

            item["models"].sort(key=lambda x: x["order"], reverse=True)
    original.sort(key=lambda x: x["order"], reverse=True)
    return original


def get_admin_url(instance: Any, admin_site: str = "admin", from_app: bool = False, **kwargs: str) -> str:
    """
    Return the admin URL for the given instance, model class or <app>.<model> string
    """
    url = "#"
    try:
        if isinstance(instance, str):
            app_label, model_name = instance.split(".")
            model_name = model_name.lower()
            url = reverse(
                "admin:{app_label}_{model_name}_changelist".format(app_label=app_label, model_name=model_name),
                current_app=admin_site,
            )

        # Model class
        elif instance.__class__ == ModelBase:
            app_label, model_name = instance._meta.app_label, instance._meta.model_name
            url = reverse(
                "admin:{app_label}_{model_name}_changelist".format(app_label=app_label, model_name=model_name),
                current_app=admin_site,
            )

        # Model instance
        elif instance.__class__.__class__ == ModelBase and isinstance(instance, instance.__class__):
            app_label, model_name = instance._meta.app_label, instance._meta.model_name
            url = reverse(
                "admin:{app_label}_{model_name}_change".format(app_label=app_label, model_name=model_name),
                args=(instance.pk,),
                current_app=admin_site,
            )

    except (NoReverseMatch, ValueError):
        # If we are not walking through the models within an app, let the user know this url cant be reversed
        if not from_app:
            logger.warning(gettext("Could not reverse url from {instance}".format(instance=instance)))

    if kwargs:
        url += "?{params}".format(params=urlencode(kwargs))

    return url


def get_filter_id(spec: ListFilter) -> str:
    return getattr(spec, "field_path", getattr(spec, "parameter_name", spec.title))


def get_custom_url(url: str, admin_site: str = "admin") -> str:
    """
    Take in a custom url, and try to reverse it
    """
    if not url:
        logger.warning("No url supplied in custom link")
        return "#"

    if "/" in url:
        return url
    try:
        url = reverse(url.lower(), current_app=admin_site)
    except NoReverseMatch:
        logger.warning("Couldnt reverse {url}".format(url=url))
        url = "#" + url

    return url


def get_model_meta(model_str: str) -> Union[None, Options]:
    """
    Get model meta class
    """
    try:
        app, model = model_str.split(".")
        model_klass: Model = apps.get_registered_model(app, model)
        return model_klass._meta
    except (ValueError, LookupError):
        return None


def get_app_admin_urls(app: str, admin_site: str = "admin") -> List[Dict]:
    """
    For the given app string, get links to all the app models admin views
    """
    if app not in apps.app_configs:
        logger.warning("{app} not found when generating links".format(app=app))
        return []

    models = []
    for model in apps.app_configs[app].get_models():
        url = get_admin_url(model, admin_site=admin_site, from_app=True)

        # We have no admin class
        if url == "#":
            continue

        models.append(
            {
                "url": url,
                "model": "{app}.{model}".format(app=model._meta.app_label, model=model._meta.model_name),
                "name": model._meta.verbose_name_plural.title(),
            }
        )

    return models


def get_view_permissions(user: AbstractUser) -> Set[str]:
    """
    Get model names based on a users view/change permissions
    """
    perms = user.get_all_permissions()
    # the perm codenames should always be lower case
    lower_perms = []
    for perm in perms:
        app, perm_codename = perm.split(".")
        lower_perms.append("{app}.{perm_codename}".format(app=app, perm_codename=perm_codename.lower()))
    return {x.replace("view_", "") for x in lower_perms if "view" in x or "change" in x}


def manage_submenu(submenus: List[Dict], options: Dict) -> List[Dict]:
    output = []
    for submenu in submenus:
        if not submenu:
            continue

        if "model" in submenu:
            model = get_model_meta(submenu["model"])
            if model:
                output.append(
                    {
                        "name": model.verbose_name_plural.title(),
                        "url": get_admin_url(submenu["model"]),
                        "icon": submenu.get("icon", options["default_icon_children"]),
                        "order": submenu.get("order", 0),
                    }
                )
        elif "url" in submenu:
            output.append(
                {
                    "name": submenu.get("name", "unspecified"),
                    "url": get_custom_url(submenu["url"]),
                    "icon": submenu.get("icon", options["default_icon_children"]),
                    "order": submenu.get("order", 0),
                }
            )
    return output


def make_single_menu(model_permissions: set[str], user: AbstractUser, link: Dict, options: Dict, app_name: Union[str, None],
                     allow_appmenus: bool = True, admin_site: str = "admin") -> List[Dict]:

    menu = []
    if isinstance(link, str):
        return menu

    perm_matches = []
    link_permissions = link.get("permissions", [])
    for perm in link_permissions:
        perm_matches.append(user.has_perm(perm))

    if not all(perm_matches):
        return menu

    # Url links
    if "url" in link:
        identifier_name = slugify(link.get("name", "unspecified"))
        menu.append(
            {
                "object_name": identifier_name,
                "name": link.get("name", "unspecified"),
                "url": get_custom_url(link["url"], admin_site=admin_site),
                "children": [],
                "new_window": link.get("new_window", False),
                "icon": link.get("icon", options["default_icon_children"]),
                "submenu": manage_submenu(link.get("submenu", []), options),
                "order": link.get("order", 0),
            }
        )
    # Model links
    elif "model" in link:
        model_label = link["model"].lower()
        app_permission = model_label
        if app_name and "." not in model_label:
            app_permission = f"{app_name}.{model_label}"
        if model_label not in model_permissions and app_permission not in model_permissions:
            return []

        _meta = get_model_meta(link["model"])

        name = _meta.verbose_name_plural.title() if _meta else link["model"]
        menu.append(
            {
                "object_name": model_label,
                "name": name,
                "url": get_admin_url(link["model"], admin_site=admin_site),
                "children": [],
                "new_window": link.get("new_window", False),
                "icon": options["icons"].get(link["model"], options["default_icon_children"]),
                "submenu": manage_submenu(link.get("submenu", []), options),
                "order": link.get("order", 0),
            }
        )

    # App links
    elif "app" in link and allow_appmenus:
        children = [
            {"name": child.get("verbose_name", child["name"]), "url": child["url"], "children": None}
            for child in get_app_admin_urls(link["app"], admin_site=admin_site)
            if child["model"] in model_permissions
        ]
        if len(children) == 0:
            return []

        menu.append({
            "object_name": link["app"],
            "name": getattr(apps.app_configs[link["app"]], "verbose_name", link["app"]).title(),
            "url": "#",
            "children": children,
            "icon": options["icons"].get(link["app"], options["default_icon_children"]),
            "order": link.get("order", 0),
        })
    return menu


def make_menu(
        user: AbstractUser, links: List[Dict], options: Dict, app_name: Union[str, None], allow_appmenus: bool = True,
        admin_site: str = "admin"
) -> List[Dict]:
    """
    Make a menu from a list of user supplied links
    """
    if not user:
        return []

    model_permissions = get_view_permissions(user)

    menu = []
    if isinstance(links, list):
        for link in links:
            menu.extend(make_single_menu(model_permissions, user, link, options, app_name, allow_appmenus, admin_site))
    elif isinstance(links, dict):
        menu.extend(make_single_menu(model_permissions, user, links, options, app_name, allow_appmenus, admin_site))

    return menu


def has_fieldsets_check(adminform: AdminForm) -> bool:
    fieldsets = adminform.fieldsets
    if not fieldsets or (len(fieldsets) == 1 and fieldsets[0][0] is None):
        return False
    return True


def attr(**kwargs) -> Callable:
    def decorator(func: Callable):
        for key, value in kwargs.items():
            setattr(func, key, value)
        return func

    return decorator


def get_installed_apps() -> List[str]:
    return [app_config.label for app_config in apps.get_app_configs()]


def hex_to_rgb(hex_color):
    """
    Convert a hex color string to an RGB string in the format "r, g, b".

    :param hex_color: str, hex color code (e.g., "#FF5733" or "FF5733")
    :return: str, "r, g, b"
    """

    # Check for valid hex color format
    try:
        if not isinstance(hex_color, str):
            raise ValueError("Hex color must be a string")

        if not hex_color.startswith('#'):
            hex_color = '#' + hex_color

        hex_color = hex_color.lstrip('#')  # Remove '#' if present
        if len(hex_color) != 6:
            raise ValueError("Invalid hex color format")

        r, g, b = (int(hex_color[i:i + 2], 16) for i in (0, 2, 4))
        return f"{r}, {g}, {b}"
    except Exception:
        return f"0, 0, 0"



def parse_date_str(value: str) -> Optional[datetime.date]:
    for date_format in settings.DATE_INPUT_FORMATS:
        try:
            return datetime.datetime.strptime(value, date_format).date()
        except (ValueError, TypeError):
            continue
    return None


def parse_datetime_str(value: str) -> Optional[datetime.datetime]:
    for date_format in settings.DATETIME_INPUT_FORMATS:
        try:
            return datetime.datetime.strptime(value, date_format)
        except (ValueError, TypeError):
            continue
    return None


def resolve_dynamic_value(value: Any, request: WSGIRequest = None, **kwargs) -> Any:
    """
    Resolve a value that can be either static or a callable function.

    Args:
        value: The value to resolve (can be static value or callable)
        request: Django request object to pass to callable
        **kwargs: Additional keyword arguments to pass to callable

    Returns:
        The resolved value
    """
    if callable(value):
        try:
            if request is not None:
                try:
                    return value(request, **kwargs)
                except TypeError as e:
                    print(f"Error calling dynamic value with request: {e}")
                    return value(**kwargs)
            else:
                return value(**kwargs)
        except Exception as e:
            print(f"Error resolving dynamic value: {e}")
            return value
    return value


def import_from_string(path: str) -> Callable:
    """
    Import a function from a dotted path string (e.g., 'myapp.utils.func')
    """
    module_path, func_name = path.rsplit('.', 1)
    module = importlib.import_module(module_path)
    return getattr(module, func_name)


def evaluate_dynamic_value(
    value: Union[str, Callable],
    *args,
    **kwargs
) -> Any:
    """
    Try to resolve a value:
    - If it's a callable, call it
    - If it's a string path to function, import and call it
    - If import/call fails, return the string as-is (assumed to be path)
    """
    if isinstance(value, (dict, list, tuple, set)):
        return value

    if callable(value):
        try:
            return value(*args, **kwargs)
        except Exception:
            ...

    if isinstance(value, str):
        try:
            func = import_from_string(value)
            return func(*args, **kwargs)
        except (ImportError, AttributeError, ValueError) as e:
            ...

    return value









from __future__ import unicode_literals

from django.apps import AppConfig

__all__ = ["DashubConfig"]


class DashubConfig(AppConfig):
    name = "dashub"
    label = "dashub"
    verbose_name = "Dashub"
